from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from ... util.enum import resolve_enum
from . import interfaces
from enum import IntEnum
from .. ledtype import LEDTYPE
# SubDriver imports
from . APA102 import APA102
from . LPD8806 import LPD8806
from . WS281X import WS281X
from . WS2801 import WS2801


# This may look weird, but was done to keep the IDs the same
# as Serial so that ledtype can always be resolved the same
SPI_DRIVERS = {
    1: LPD8806,
    2: WS2801,
    3: WS281X,
    9: APA102
}


def SPI(ledtype, num, **kwargs):
    """Wrapper function for using SPI device drivers on systems like the Raspberry Pi and BeagleBone"""
    ledtype = resolve_enum(LEDTYPE, ledtype)
    if ledtype not in SPI_DRIVERS.keys():
        raise ValueError('{} is not a valid LED type.'.format(type))

    return SPI_DRIVERS[ledtype](num, **kwargs)



import copy

from ..util.datafile import DataFile

"""
The BP preset library is a JSON file with five named sections:

    project, driver, layout, animation, run.

Each section is a dictionary that maps a name to a dictionary called a "preset".

For each section, a special preset named "default" is used if none is specified,
or the system default is used if no preset named "default" exists.

If a library preset has a key "extends" then it extends and overrides
another setting in the same section.
"""


def get_setting(section, name):
    extensions = []

    while name:
        setting = section[name]
        extensions.append(setting)
        name = setting.get('extends')

    setting = {}
    for s in reversed(extensions):
        setting.update(s)

    try:
        del setting['extends']
    except KeyError:
        pass
    return setting


class PresetLibrary(DataFile):
    def _section(self, section):
        return self.data.setdefault(section, {})

    def get_setting(self, section, name):
        return get_setting(self._section(section), name)

    def get_value(self, section, name, key):
        return self.get_setting(section, name)[key]

    def set_setting(self, section, name, setting):
        self._section(section)[name] = setting
        self.write()

    def set_value(self, section, name, key, value):
        self._section(section).set_default(name, {})[key] = value
        self.write()



import gitty, json, os, sys
from . import defaults, importer
from . types.defaults import FIELD_TYPES
from .. animation import runner
from .. import data_maker
from .. layout.geometry import gen_matrix
from .. layout.multimap import MultiMapBuilder
from .. util import files
from .. drivers.serial import codes


def _make_object(*args, field_types=FIELD_TYPES, **kwds):
    return importer.make_object(*args, field_types=field_types, **kwds)


def _make_layout(driver, layout, maker=None):
    maker = data_maker.Maker(**(maker or {}))
    drivers = []

    def multi_drivers(device_ids, width, height, serpentine=False, **kwds):
        build = MultiMapBuilder()

        for id in device_ids:
            build.addRow(gen_matrix(width, height, serpentine=serpentine))
            d = _make_object(width=width, height=height, deviceID=id, **kwds)
            drivers.append(d)

        return build.map

    def make_drivers(multimap=False, **kwds):
        if multimap:
            return multi_drivers(**kwds)

        drivers.append(_make_object(**kwds))

    coordMap = make_drivers(maker=maker, **driver)
    return _make_object(drivers, coordMap=coordMap, maker=maker, **layout)


def make_animation(layout, animation, run=None):
    animation = _make_object(layout, **animation)
    animation.set_runner(runner.Runner(**(run or {})))
    return animation


def _make_project(path=None, **project):
    gitty.sys_path.extend(path)

    kwds = defaults.apply_defaults(project)
    animation = kwds.pop('animation', {})
    run = kwds.pop('run', {})
    layout = _make_layout(**kwds)
    return make_animation(layout, animation, run)


def _apply_defaults(project, defaults):
    # TODO: consolidate with bp.project.defaults.
    for k, v in defaults.items():
        if k == 'ledtype' or not v:
            continue

        if k not in project:
            project[k] = v

        else:
            existing = project[k]
            if 'typename' not in existing:
                try:
                    existing['typename'] = v
                except:
                    pass

    ledtype = defaults.get('ledtype')
    if ledtype:
        if 'ledtype' not in project['driver']:
            project['driver']['ledtype'] = ledtype


def project_to_animation(project, defaults=None):
    _apply_defaults(project, defaults or {})
    return _make_project(**project)



def field_value(field, value, field_types):
    if not (field_types and field in field_types):
        return value

    try:
        return field_types[field].make(value)
    except Exception as e:
        e.args += ('for field %s and value %s' % (field, value), )
        raise


def component(comp, field_types):
    return {k: field_value(k, v, field_types) for k, v in comp.items()}


def project(proj, field_types):
    return {k: component(v, field_types) for k, v in proj.items()}



# Original code by msurguy: https://github.com/ManiacalLabs/BiblioPixel/issues/51#issuecomment-228662943

from . channel_order import ChannelOrder
from . driver_base import DriverBase
from .. import log
from .. import gamma

try:
    from neopixel import Adafruit_NeoPixel, Color as NeoColor
except:
    WS_ERROR = """PiWS281X Requires the rpi_ws281x C extension.
Install rpi_ws281x with below commands:
git clone https://github.com/jgarff/rpi_ws281x.git
cd rpi_ws281x
sudo apt-get install python-dev swig scons
sudo scons
cd python
# If using default system python3
sudo python3 setup.py build install
# If using virtualenv, enter env then run
python setup.py build install
    """
    log.error(WS_ERROR)
    raise


class PiWS281X(DriverBase):
    """
    Driver for controlling WS281X LEDs via the rpi_ws281x C-extension.
    Only supported on the Raspberry Pi 2 & 3
    """
    def __init__(self, num, gamma=gamma.NEOPIXEL, c_order=ChannelOrder.RGB, gpio=18,
                 ledFreqHz=800000, ledDma=5, ledInvert=False, **kwds):
        """
        num - Number of LED pixels.
        gpio - GPIO pin connected to the pixels (must support PWM! GPIO 13 or 18 (pins 33 or 12) on RPi 3).
        ledFreqHz - LED signal frequency in hertz (800khz or 400khz)
        ledDma - DMA channel to use for generating signal (Between 1 and 14)
        ledInvert - True to invert the signal (when using NPN transistor level shift)
        """
        super().__init__(num, c_order=c_order, gamma=gamma, **kwds)
        self.gamma = gamma
        self._strip = Adafruit_NeoPixel(num, gpio, ledFreqHz,
                                        ledDma, ledInvert, 255, 0, 0x081000)
        # Intialize the library (must be called once before other functions).
        self._strip.begin()

    def set_brightness(self, brightness):
        self._strip.setBrightness(brightness)
        return True

    def _compute_packet(self):
        self._render()
        data = self._buf
        self._packet = [tuple(data[(p * 3):(p * 3) + 3]) for p in range(len(data) // 3)]

    def _send_packet(self):
        for i, p in enumerate(self._packet):
            self._strip.setPixelColor(i, NeoColor(*p))

        self._strip.show()



import ctypes
from multiprocessing.sharedctypes import RawArray

# Note https://stackoverflow.com/questions/37705974/

BYTE, FLOAT = ctypes.c_uint8, ctypes.c_float


def shared_list_maker(type):
    return lambda size: RawArray(type, size)


def list_maker(size):
    return [(0, 0, 0)] * size


class Maker(object):
    def __init__(self, integer=False, shared_memory=False):
        if shared_memory:
            self.make_packet = shared_list_maker(BYTE)
            number_type = BYTE if integer else FLOAT
            self.color_list = shared_list_maker(3 * number_type)

        else:
            self.make_packet = bytearray
            self.color_list = list_maker


MAKER = Maker()
ColorList = MAKER.color_list



import copy
from .importer import import_symbol

ALIASES = {
    'driver': {
        'apa102': 'bibliopixel.drivers.API.APA102.APA102',
        'sk9822': 'bibliopixel.drivers.API.APA102.APA102',
        'dummy': 'bibliopixel.drivers.dummy_driver.Dummy',
        'hue': 'bibliopixel.drivers.hue.Hue',
        'image': 'bibliopixel.drivers.image_sequence.ImageSequence',
        'lpd8806': 'bibliopixel.drivers.API.LPD8806.LPD8806',
        'network': 'bibliopixel.drivers.network.Network',
        'network_udp': 'bibliopixel.drivers.network.NetworkUDP',
        'serial': 'bibliopixel.drivers.serial.Serial',
        'simpixel': 'bibliopixel.drivers.SimPixel.SimPixel',
        'ws281x': 'bibliopixel.drivers.API.WS281X.WS281X',
        'ws2801': 'bibliopixel.drivers.API.WS2801.WS2801',
        'spi': 'bibliopixel.drivers.SPI.SPI',
        'pi_ws281x': 'bibliopixel.drivers.PiWS281X.PiWS281X'
    },

    'layout': {
        'circle': 'bibliopixel.layout.circle.Circle',
        'cube': 'bibliopixel.layout.cube.Cube',
        'matrix': 'bibliopixel.layout.matrix.Matrix',
        'pov': 'bibliopixel.layout.pov.POV',
        'strip': 'bibliopixel.layout.strip.Strip',
    },

    'animation': {
        'off': 'bibliopixel.animation.off.OffAnim',
        'matrix_calibration':
        'bibliopixel.animation.tests.MatrixCalibrationTest',
        'matrix_test': 'bibliopixel.animation.tests.MatrixChannelTest',
        'receiver': 'bibliopixel.animation.receiver.BaseReceiver',
        'sequence': 'bibliopixel.animation.Sequence',
        'strip_test': 'bibliopixel.animation.tests.StripChannelTest',
    },
}


def fill_typename(desc, key):
    if isinstance(desc, str):
        return fill_typename({'typename': desc}, key)

    typename = desc.get('typename')
    if typename:
        desc['typename'] = ALIASES[key].get(typename, typename)

    return desc


def resolve_aliases(project):
    result = copy.deepcopy(project)
    for key in ALIASES:
        if key in result:
            result[key] = fill_typename(result[key], key)
    return result



import argparse, gitty, os, sys
from .. import log
from .. project.importer import import_symbol
from .. project.preset_library import PresetLibrary

__all__ = ['main']

COMMANDS = ('all_pixel', 'clear_cache', 'devices', 'demo', 'run')
MODULES = {c: import_symbol('.' + c, 'bibliopixel.main') for c in COMMANDS}
PRESET_LIBRARY_DEFAULT = '~/.bibliopixel'
LOG_LEVELS = ('debug', 'info', 'warning', 'error', 'critical')
ENABLE_PRESETS = False

PATH_HELP = """\
A list of directories, separated by colons, 'which are added to the end of
`sys.path`.

You can also use gitty-style paths which start with `//git/` to
dynamically load a library from a public git repository.

See https://github.com/ManiacalLabs/BiblioPixel/wiki/BiblioPixel-Paths
for more information.
"""

LOGLEVEL_HELP = """\
Set what level of events to log. Higher log levels print less."""


def no_command(*_):
    print('ERROR: No command entered')
    print('Valid:', ', '.join(COMMANDS))
    return -1


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    for name, module in sorted(MODULES.items()):
        subparser = subparsers.add_parser(name, help=module.HELP)
        module.set_parser(subparser)
    parser.add_argument('--loglevel', choices=LOG_LEVELS, default='info',
                        help=LOGLEVEL_HELP)
    parser.add_argument('--path', default=None,
                        help=PATH_HELP)
    if ENABLE_PRESETS:
        parser.add_argument('--presets', help='Filename for presets',
                            default=PRESET_LIBRARY_DEFAULT)

    args = ['--help' if i == 'help' else i for i in sys.argv[1:]]
    args = parser.parse_args(args)

    log.set_log_level(args.loglevel)
    presets = ENABLE_PRESETS and PresetLibrary(
        os.path.expanduser(args.presets), True)

    run = getattr(args, 'run', no_command)
    gitty.sys_path.extend(args.path)
    sys.exit(run(args, presets) or 0)



class Rotation:
    ROTATE_0 = 0  # no rotation
    ROTATE_90 = 3  # rotate 90 degrees
    ROTATE_180 = 2  # rotate 180 degrees
    ROTATE_270 = 1  # rotate 270 degrees


def rotate_and_flip(coord_map, rotation, flip):
    for i in range(rotation):
        coord_map = list(zip(*coord_map[::-1]))

    if flip:
        coord_map = coord_map[::-1]

    return coord_map



from . animation import BaseAnimation
from .. layout import Cube


class BaseCubeAnim(BaseAnimation):

    def __init__(self, layout):
        super().__init__(layout)
        if not isinstance(layout, Cube):  # pragma: no cover
            raise RuntimeError('Must use bibliopixel.layout.Cube with ' +
                               'Cube Animations!')

        self.x, self.y, self.z = layout.x, layout.y, layout.z



from ..project.defaults import SECTIONS

HELP = """
Show bibliopixel settings.
"""


def run(args, settings):
    def show_all():
        if not settings.data:
            print('No settings.')
            return

        first = True
        for name, section in sorted(settings.data.items()):
            if first:
                first = False
            else:
                print()

            show_section(name, section)

    def show_section(name, section):
        print('%s:' % name)
        for i in sorted(section.keys()):
            print('    %s' % i)

    def show_setting(setting):
        for k, v in sorted(setting.items()):
            print('%s = %s' % (k, v))

    if not args.name:
        show_all()
        return

    parts = args.name.split('.')
    section_name = parts.pop(0)

    try:
        section = settings.data[section_name]
    except KeyError:
        print('ERROR: no section %s' % section_name)
        return -1

    if not parts:
        show_section(section_name, section)
        return

    name = parts.pop(0)
    try:
        setting = section[name]
    except KeyError:
        print('ERROR: no setting %s in section %s' % (section_name, section))
        return -1

    if not parts:
        show_setting(setting)
        return

    key = parts.pop(0)
    if parts:
        print('ERROR: too many segments in name %s' % args.name)
        return -1

    try:
        value = setting[args.key]
    except:
        print('ERROR: no key %s in setting %s in section %s' %
              (key, setting, section_name))
        return -1

    print(value)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.add_argument('name', nargs='?', default='')



import json
from ..project.defaults import SECTIONS

HELP = """
Set bibliopixel values
"""


def run(args, settings):
    parts = args.name.split('.')

    if len(parts) == 2:
        parts.append('')

    elif len(parts) != 3:
        print('ERROR: name must have 2 or 3 parts: %s' % args.name)
        return -1

    section, name, key = parts

    if section not in SECTIONS:
        print('ERROR: no section %s' % section)
        return -1

    sdict = settings.data.setdefault(section, {})

    if key:
        sdict.setdefault(name, {})[key] = args.value
    else:
        try:
            value = json.loads(args.value)
        except:
            print('ERROR: Can\'t parse json value: %s' % args.value)
            return -1

        if not isinstance(value, dict):
            print('ERROR: value must be a dictionary: %s' % args.value)
            return -1

        sdict[name] = value

    settings.write()


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.add_argument('name')
    parser.add_argument('value')



from . import channel_order, color, duration, gamma, ledtype

FIELD_TYPES = {
    'c_order': channel_order,
    'color': color,
    'duration': duration,
    'gamma': gamma,
    'ledtype': ledtype,
    'time': duration,
    'ledtype': ledtype,
}



import json
from .. project import project
from .. import log
from .. util import files

HELP = """
Run a project description file.
"""


def add_project_default_arguments(parser):
    parser.add_argument(
        '-d', '--driver', default=None,
        help='Default driver type if no driver is specified')

    parser.add_argument(
        '-l', '--layout', default=None,
        help='Default LAYOUT class if no LAYOUT is specified')

    parser.add_argument(
        '-t', '--ledtype', default=None,
        help='Default LED type if no LED type is specified')

    parser.add_argument(
        '-a', '--animation', default=None,
        help='Default animation type if no animation is specified')


def get_project_default_arguments(args):
    return {
        'driver': args.driver,
        'layout': args.layout,
        'animation': args.animation,
        'ledtype': args.ledtype,
    }


def project_to_animation(name, is_json, defaults):
    log.info('Processing project file...')
    data = name if is_json else files.opener(name).read()
    try:
        desc = json.loads(data)
    except ValueError as e:
        e.args += tuple(['in %s' % name])
        raise
    return project.project_to_animation(desc, defaults)


def run(args, settings):
    if args.name:
        defaults = get_project_default_arguments(args)
        project_to_animation(args.name, args.json, defaults).start()


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = 'Run specified BiblioPixel project from file or URL.'

    parser.add_argument(
        'name', nargs='?',
        help='Path project file - can be a URL or file system location',
        default='')

    parser.add_argument(
        '-j', '--json', action='store_true',
        help='Enter JSON directly as a command line argument.')

    add_project_default_arguments(parser)



from . import index_ops
from . import Rotation, rotate_and_flip


class Matrix(object):
    """Map a matrix onto a strip of lights."""

    def __init__(self, strip, columns=None, rows=None,
                 reflect_x=False, reflect_y=False,
                 serpentine_x=False, serpentine_y=False,
                 transpose=False):
        self.strip = strip

        if columns:
            if not rows:
                rows = len(strip) // columns
            elif rows * columns > len(strip):
                raise IndexError('Index out of range.')
        elif rows:
            columns = len(strip) // rows
        else:
            raise ValueError('Must supply either rows or columns')
        self.columns, self.rows = columns, rows

        ops = (reflect_x and index_ops.reflect_x,
               reflect_y and index_ops.reflect_y,
               serpentine_x and index_ops.serpentine_x,
               serpentine_y and index_ops.serpentine_y,
               transpose and index_ops.transpose)
        self.operations = list(filter(None, ops))

    def _index(self, x, y):
        for o in self.operations:
            x, y = o(x, y, self)

        if 0 <= x < self.columns and 0 <= y < self.rows:
            return x + y * self.columns

        raise IndexError('Index out of range.')

    def get(self, x, y):
        return self.strip[self._index(x, y)]

    def set(self, x, y, value):
        self.strip[self._index(x, y)] = value


def gen_matrix(dx, dy, serpentine=True, offset=0,
               rotation=Rotation.ROTATE_0, y_flip=False):
    """Helper method to generate X,Y coordinate maps for strips"""

    result = []
    for y in range(dy):
        if not serpentine or y % 2 == 0:
            result.append([(dx * y) + x + offset for x in range(dx)])
        else:
            result.append([((dx * (y + 1)) - 1) - x + offset for x in range(dx)])

    result = rotate_and_flip(result, rotation, y_flip)

    return result


def pixel_positions_from_matrix(coord_map):
    max_width = 0
    for x in coord_map:
        if len(x) > max_width:
            max_width = len(x)

    num = len(coord_map) * max_width
    result = [None] * num

    for y in range(len(coord_map)):
        for x in range(len(coord_map[y])):
            result[coord_map[y][x]] = [x, y, 0]

    return result



import contextlib, threading, time
from . runner import Runner
from .. import log
from .. threads.animation_threading import AnimationThreading


class BaseAnimation(object):
    free_run = False

    def __init__(self, layout):
        self.layout = layout
        self.internal_delay = None

    @property
    def _led(self):
        """Many BiblioPixelAnimations use the "protected" variable _led."""
        return self.layout

    @_led.setter
    def _led(self, layout):
        self.layout = layout

    def preRun(self, amt=1):
        self.layout.all_off()

    def step(self, amt=1):
        raise RuntimeError("Base class step() called. This shouldn't happen")

    def cleanup(self):
        self.threading.stop_thread(wait=True)
        self.layout.cleanup()

    def is_running(self):
        if self.threading.stop_event.isSet():
            return False

        if self.runner.max_steps:
            return self.cur_step < self.runner.max_steps

        return not (self.runner.until_complete and self.completed)

    def run_one_frame(self):
        timestamps = []

        def stamp():
            timestamps.append(time.time())

        stamp()

        self.step(self.runner.amt)

        stamp()

        self.layout.frame_render_time = timestamps[1] - timestamps[0]
        self.layout.push_to_driver()

        stamp()

        _report_framerate(timestamps)

        self.cur_step += 1
        if self.completed and self.runner.max_cycles > 0:
            if self.cycle_count < self.runner.max_cycles - 1:
                self.cycle_count += 1
                self.completed = False

        stamp()

        self.threading.wait(self.sleep_time, timestamps)

    @contextlib.contextmanager
    def run_context(self):
        self.completed = False
        self._step = 0
        self.cur_step = 0
        self.cycle_count = 0

        if self.free_run:
            self.sleep_time = None
        elif self.internal_delay:
            self.sleep_time = self.internal_delay
        else:
            self.sleep_time = self.runner.sleep_time
        self.layout.animation_sleep_time = self.sleep_time or 0

        self.preRun(self.runner.amt)
        try:
            yield
        finally:
            self.cleanup()

    def run_all_frames(self):
        with self.run_context():
            while self.is_running():
                self.run_one_frame()

    def set_runner(self, runner):
        self.runner = runner
        self.threading = AnimationThreading(self.runner, self.run_all_frames)

    def start(self):
        self.threading.start()

    def run(self, **kwds):
        # DEPRECATED
        self.set_runner(Runner(**kwds))
        self.start()


def _report_framerate(timestamps):
    total_time = timestamps[-1] - timestamps[0]
    fps = int(1.0 / max(total_time, 0.001))
    log.debug("%dms/%dfps / Frame: %dms / Update: %dms",
              1000 * total_time,
              fps,
              1000 * (timestamps[1] - timestamps[0]),
              1000 * (timestamps[2] - timestamps[1]))



from .. project import aliases, project
from . import animation, runner


class Sequence(animation.BaseAnimation):
    def __init__(self, layout, animations=None):
        def make_animation(a):
            if isinstance(a, str):
                desc = {'animation': a[0]}
            elif isinstance(a, dict):
                desc = a
            else:
                desc = {'animation': a[0], 'run': a[1]}
            desc = aliases.resolve_aliases(desc)
            return project.make_animation(layout=layout, **desc)

        super().__init__(layout)
        self.animations = [make_animation(i) for i in animations or []]
        self.index = 0
        self.internal_delay = 0  # never wait

    # overriding to handle all the animations
    def stopThread(self, wait=False):
        for a, r in self.animations:
            # a bit of a hack. they aren't threaded, but stops them anyway
            a.thread_strategy.stop_event.set()
        self.thread_strategy.stop_thread(wait)

    def add_animation(self, anim, **kwds):
        # DEPRECATED.
        anim.set_runner(runner.Runner(**kwds))
        self.animations.append(anim)

    def preRun(self, amt=1):
        self.index = -1

    @property
    def current_animation(self):
        return self.animations[self.index]

    def step(self, amt=1):
        self.index += 1
        if self.index >= len(self.animations):
            if self.runner.until_complete:
                self.completed = True
            else:
                self.index = 0

        if not self.completed and self.animations:
            self.current_animation.run_all_frames()



import importlib


def import_symbol(typename, package=None):
    """Import a module or typename within a module from its name."""
    try:
        return importlib.import_module(typename, package=package)

    except ImportError as e:
        parts = typename.split('.')
        if len(parts) > 1:
            typename = parts.pop()

            # Call import_module recursively.
            namespace = import_symbol('.'.join(parts), package=package)

            try:
                return getattr(namespace, typename)
            except AttributeError:
                pass
        raise



import math, threading, time

from .. import colors, data_maker, font, log, matrix
from . layout import Layout
from . geometry.cube import gen_cube, pixel_positions_from_cube


class Cube(Layout):

    def __init__(self, drivers, x, y, z, coordMap=None,
                 threadedUpdate=False, brightness=255, **kwargs):
        super().__init__(drivers, threadedUpdate, brightness,
                         maker=kwargs.get('maker', data_maker.MAKER))

        self.x = x
        self.y = y
        self.z = z

        if self.x * self.y * self.z != self.numLEDs:
            raise TypeError("(x * y * z) MUST equal the total pixel count!")

        if coordMap:
            self.cube_map = coordMap
        else:
            if len(self.drivers) == 1:
                log.info('Auto generating coordinate map. Use gen_cube directly if more control needed.')
                self.cube_map = gen_cube(x, y, z)
            else:
                raise TypeError("Must provide coordMap if using multiple drivers!")

        self.set_pixel_positions(pixel_positions_from_cube(self.cube_map))

    def get_pixel_positions(self):
        return pixel_positions_from_cube(self.coord_map)

    def set(self, x, y, z, color):
        try:
            pixel = self.cube_map[z][y][x]
            self._set_base(pixel, color)
        except IndexError:
            pass

    def get(self, x, y, z):
        try:
            pixel = self.cube_map[z][y][x]
            return self._get_base(pixel)
        except IndexError:
            return 0, 0, 0

    def setHSV(self, x, y, z, hsv):
        color = colors.hsv2rgb(hsv)
        self._set(x, y, z, color)

    def setRGB(self, x, y, z, r, g, b):
        color = (r, g, b)
        self._set(x, y, z, color)


# This is DEPRECATED
LEDCube = Cube



class Runner(object):

    def __init__(self, amt=1, fps=None, sleep_time=0, max_steps=0,
                 until_complete=False, max_cycles=0, seconds=None,
                 threaded=False, join_thread=False):
        assert max_steps >= 0
        assert sleep_time >= 0
        assert max_cycles >= 0
        assert not (sleep_time and fps)
        assert not (seconds and max_steps)

        self.amt = amt
        self.fps = fps
        self.sleep_time = 1 / fps if fps else sleep_time
        self.until_complete = until_complete
        if seconds is None:
            self.max_steps = max_steps
        else:
            self.max_steps = int(seconds / self.sleep_time)
        self.max_cycles = max_cycles
        self.seconds = seconds
        self.threaded = threaded
        self.join_thread = join_thread



from . animation import BaseAnimation
from .. layout import Matrix


class BaseMatrixAnim(BaseAnimation):

    def __init__(self, layout, width=0, height=0, startX=0, startY=0):
        super().__init__(layout)
        if not isinstance(layout, Matrix):  # pragma: no cover
            raise RuntimeError('Must use bibliopixel.layout.Matrix with ' +
                               'Matrix Animations!')

        self.width = width or layout.width
        self.height = height or layout.height

        self.startX = startX
        self.startY = startY



import json
from . import files


class DataFile(object):
    def __init__(self, filename, open=files.opener):
        self.filename = filename

        self.open = open
        self.data = {}
        self.read()

    def read(self):
        try:
            fp = self.open(self.filename)
        except FileNotFoundError:
            self.data = {}
        else:
            self.data = json.load(fp)

    def write(self):
        json.dump(self.data, self.open(self.filename, 'w'))
